/*!
 * gifloop v0.1.0
  */
(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports) :
  typeof define === 'function' && define.amd ? define(['exports'], factory) :
  (global = global || self, factory(global.gifloop = global.gifloop || {}));
}(this, function (exports) { 'use strict';

  function mitt(n){return {all:n=n||new Map,on:function(t,e){var i=n.get(t);i?i.push(e):n.set(t,[e]);},off:function(t,e){var i=n.get(t);i&&(e?i.splice(i.indexOf(e)>>>0,1):n.set(t,[]));},emit:function(t,e){var i=n.get(t);i&&i.slice().map(function(n){n(e);}),(i=n.get("*"))&&i.slice().map(function(n){n(t,e);});}}}

  var __defProp = Object.defineProperty;
  var __defProps = Object.defineProperties;
  var __getOwnPropDescs = Object.getOwnPropertyDescriptors;
  var __getOwnPropSymbols = Object.getOwnPropertySymbols;
  var __hasOwnProp = Object.prototype.hasOwnProperty;
  var __propIsEnum = Object.prototype.propertyIsEnumerable;
  var __defNormalProp = (obj, key, value) => key in obj ? __defProp(obj, key, { enumerable: true, configurable: true, writable: true, value }) : obj[key] = value;
  var __spreadValues = (a, b) => {
    for (var prop in b || (b = {}))
      if (__hasOwnProp.call(b, prop))
        __defNormalProp(a, prop, b[prop]);
    if (__getOwnPropSymbols)
      for (var prop of __getOwnPropSymbols(b)) {
        if (__propIsEnum.call(b, prop))
          __defNormalProp(a, prop, b[prop]);
      }
    return a;
  };
  var __spreadProps = (a, b) => __defProps(a, __getOwnPropDescs(b));
  var __publicField = (obj, key, value) => {
    __defNormalProp(obj, typeof key !== "symbol" ? key + "" : key, value);
    return value;
  };
  class Stream {
    constructor(arrayBuffer, littleEndian) {
      __publicField(this, "index", 0);
      __publicField(this, "dataView");
      this.arrayBuffer = arrayBuffer;
      this.littleEndian = littleEndian;
      this.dataView = new DataView(arrayBuffer);
      this.littleEndian = littleEndian || false;
    }
    check(size) {
      if (this.index + size <= this.dataView.byteLength) {
        return true;
      }
      return false;
    }
    seek(x) {
      this.index += x;
    }
    peekByte(offset = 0) {
      this.check(1 + offset);
      return this.dataView.getUint8(this.index + offset);
    }
    slice(offset, length) {
      const ab = new Uint8Array(this.dataView.buffer, offset, length);
      return ab;
    }
    readUint8() {
      this.check(1);
      const value = this.dataView.getUint8(this.index);
      this.seek(1);
      return value;
    }
    readUint8Array(len) {
      const r = [];
      for (let i = 0; i < len; i++) {
        r.push(this.readUint8());
      }
      return r;
    }
    readUint16() {
      this.check(2);
      const value = this.dataView.getUint16(this.index, this.littleEndian);
      this.seek(2);
      return value;
    }
    readUint32() {
      this.check(4);
      const val = this.dataView.getUint32(this.index, this.littleEndian);
      this.seek(4);
      return val;
    }
    readInt32() {
      this.check(4);
      const val = this.dataView.getInt32(this.index, this.littleEndian);
      this.seek(4);
      return val;
    }
    readString(len) {
      let str = "";
      for (let i = 0; i < len; i++) {
        str += String.fromCharCode(this.readUint8());
      }
      return str;
    }
    readEscapedString(len) {
      let str = "";
      for (let i = 0; i < len; i++) {
        const char = this.readUint8();
        if (char >= 32 && char <= 126) {
          str += String.fromCharCode(char);
        } else {
          str += `\\x${char.toString(16).padStart(2, "0")}`;
        }
      }
      return str;
    }
    readNullTerminatedString() {
      let str = "";
      let val = this.readUint8();
      while (val !== 0) {
        str += String.fromCharCode(val);
        val = this.readUint8();
      }
      return str;
    }
    hasMore(x) {
      this.check(1);
      return this.index <= this.dataView.byteLength - (x || 1);
    }
    setOffset(offset) {
      this.index = offset;
    }
    getOffset() {
      return this.index;
    }
    readInt8() {
      const value = this.dataView.getInt8(this.index);
      this.seek(1);
      return value;
    }
  }
  function bitsToNumber(ba) {
    return ba.reduce((s, n) => s * 2 + n, 0);
  }
  function byteToBits(bite) {
    let a = [];
    for (let i = 7; i >= 0; i--) {
      a.push(!!(bite & 1 << i));
    }
    return a.map((n) => Number(n));
  }
  function getBytes(stream, offset, length) {
    try {
      return stream.slice(offset, length);
    } catch (error) {
      return new Uint8Array(0);
    }
  }
  function numberToHex(n) {
    const hex = parseInt(`${n}`).toString(16);
    return (hex.length === 2 ? hex : `0${hex}`).toUpperCase();
  }
  class Header {
    constructor(stream) {
      __publicField(this, "type", "Header");
      __publicField(this, "offset", 0);
      __publicField(this, "length", 0);
      __publicField(this, "bytes", new Uint8Array(0));
      __publicField(this, "signature", "");
      __publicField(this, "version", "");
      this.stream = stream;
      this.stream = stream;
    }
    parse() {
      this.offset = this.stream.getOffset();
      this.signature = this.stream.readString(3);
      this.version = this.stream.readString(3);
      this.length = this.stream.getOffset() - this.offset;
      if (this.signature !== "GIF")
        throw new Error("Not a GIF file.");
    }
    export() {
      return {
        type: this.type,
        offset: this.offset,
        length: this.length,
        bytes: getBytes(this.stream, this.offset, this.length),
        data: {
          signature: this.signature,
          version: this.version
        }
      };
    }
  }
  class LogicalScreenDescriptor {
    constructor(stream) {
      __publicField(this, "type", "LogicalScreenDescriptor");
      __publicField(this, "offset", 0);
      __publicField(this, "length", 0);
      __publicField(this, "bytes", new Uint8Array(0));
      __publicField(this, "width", 0);
      __publicField(this, "height", 0);
      __publicField(this, "packedFields", []);
      __publicField(this, "globalColorTableFlag", 0);
      __publicField(this, "colorResolution", 0);
      __publicField(this, "sortFlag", 0);
      __publicField(this, "globalColorTableSize", 0);
      __publicField(this, "backgroundColorIndex", 0);
      __publicField(this, "pixelAspectRatio", 0);
      this.stream = stream;
      this.stream = stream;
    }
    parse() {
      this.offset = this.stream.getOffset();
      this.width = this.stream.readUint16();
      this.height = this.stream.readUint16();
      this.packedFields = byteToBits(this.stream.readUint8());
      this.handlerPackedField(this.packedFields);
      this.backgroundColorIndex = this.stream.readUint8();
      this.pixelAspectRatio = this.stream.readInt8();
      this.length = this.stream.getOffset() - this.offset;
    }
    hasGlobalColorTable() {
      return !!this.globalColorTableFlag;
    }
    getGlobalColorTableSize() {
      return this.globalColorTableSize;
    }
    export() {
      return {
        type: this.type,
        offset: this.offset,
        length: this.length,
        bytes: getBytes(this.stream, this.offset, this.length),
        data: {
          width: this.width,
          height: this.height,
          packedFields: this.packedFields,
          globalColorTableFlag: this.globalColorTableFlag,
          colorResolution: this.colorResolution,
          sortFlag: this.sortFlag,
          globalColorTableSize: this.globalColorTableSize,
          backgroundColorIndex: this.backgroundColorIndex,
          pixelAspectRatio: this.pixelAspectRatio
        }
      };
    }
    handlerPackedField(rb) {
      const bits = [...rb];
      this.globalColorTableFlag = bits.shift() || 0;
      this.colorResolution = bitsToNumber(bits.splice(0, 3)) + 1;
      this.sortFlag = bits.shift() || 0;
      this.globalColorTableSize = bitsToNumber(bits.splice(0, 3));
    }
  }
  class ColorTable {
    constructor(stream, type) {
      __publicField(this, "type", "");
      __publicField(this, "offset", 0);
      __publicField(this, "length", 0);
      __publicField(this, "bytes", new Uint8Array(0));
      __publicField(this, "colors", []);
      this.stream = stream;
      this.stream = stream;
      this.type = type;
    }
    parse({ size = 0 }) {
      const globalColorTableSize = 1 << size + 1;
      this.offset = this.stream.getOffset();
      for (let i = 0; i < globalColorTableSize; i++) {
        this.colors.push(this.stream.readUint8Array(3));
      }
      this.length = this.stream.getOffset() - this.offset;
    }
    export() {
      return {
        type: this.type,
        offset: this.offset,
        length: this.length,
        bytes: getBytes(this.stream, this.offset, this.length),
        data: this.colors
      };
    }
  }
  class GlobalColorTable extends ColorTable {
    constructor(stream) {
      super(stream, "GlobalColorTable");
    }
  }
  class BaseExtension {
    constructor(stream) {
      __publicField(this, "stream");
      this.stream = stream;
    }
    parse(param) {
    }
    export() {
    }
    readSubBlocks() {
      const offset = this.stream.getOffset();
      let size = this.stream.readInt8();
      let length = 1;
      const blocks = [];
      while (size > 0) {
        blocks.push({ offset: offset + length, length: size });
        length += size + 1;
        this.stream.seek(size);
        size = this.stream.readUint8();
      }
      return {
        blocks,
        blocksLength: length
      };
    }
  }
  class GraphicsControlExtension extends BaseExtension {
    constructor(stream) {
      super(stream);
      __publicField(this, "type", "GraphicsControlExtension");
      __publicField(this, "offset", 0);
      __publicField(this, "length", 0);
      __publicField(this, "bytes", new Uint8Array(0));
      __publicField(this, "introducer", "");
      __publicField(this, "label", "");
      __publicField(this, "blockSize", 0);
      __publicField(this, "packedFields", []);
      __publicField(this, "reserved", []);
      __publicField(this, "disposalMethod", 0);
      __publicField(this, "userInputFlag", 0);
      __publicField(this, "transparentColorFlag", 0);
      __publicField(this, "delayTime", 0);
      __publicField(this, "transparentColorIndex", 0);
      __publicField(this, "blockTerminator", "");
      this.stream = stream;
    }
    parse({ introducer = 0, label = 0 }) {
      this.offset = this.stream.getOffset() - 2;
      this.introducer = String.fromCharCode(introducer);
      this.label = numberToHex(label);
      this.blockSize = this.stream.readUint8();
      this.packedFields = byteToBits(this.stream.readUint8());
      const bits = [...this.packedFields];
      this.reserved = bits.splice(0, 3);
      this.disposalMethod = bitsToNumber(bits.splice(0, 3));
      this.userInputFlag = bits.shift() || 0;
      this.transparentColorFlag = bits.shift() || 0;
      this.delayTime = this.stream.readUint16();
      this.transparentColorIndex = this.stream.readUint8();
      this.blockTerminator = this.stream.readString(1);
      this.length = this.stream.getOffset() - this.offset;
    }
    export() {
      return {
        type: this.type,
        offset: this.offset,
        length: this.length,
        bytes: getBytes(this.stream, this.offset, this.length),
        data: {
          introducer: this.introducer,
          label: this.label,
          blockSize: this.blockSize,
          packedFields: this.packedFields,
          reserved: this.reserved,
          disposalMethod: this.disposalMethod,
          userInputFlag: this.userInputFlag,
          transparentColorFlag: this.transparentColorFlag,
          delayTime: this.delayTime,
          transparentColorIndex: this.transparentColorIndex,
          blockTerminator: this.blockTerminator
        }
      };
    }
  }
  class CommentExtension extends BaseExtension {
    constructor(stream) {
      super(stream);
      __publicField(this, "type", "CommentExtension");
      __publicField(this, "offset", 0);
      __publicField(this, "length", 0);
      __publicField(this, "bytes", new Uint8Array(0));
      __publicField(this, "comment", {});
      this.stream = stream;
    }
    parse({ introducer, label }) {
      this.offset = this.stream.getOffset();
      this.comment = this.readSubBlocks();
      this.length = this.stream.getOffset() - this.offset;
    }
    export() {
      return {
        type: this.type,
        offset: this.offset,
        length: this.length,
        bytes: getBytes(this.stream, this.offset, this.length),
        data: this.comment
      };
    }
  }
  class PlainTextExtension extends BaseExtension {
    constructor(stream) {
      super(stream);
      __publicField(this, "type", "PlainTextExtension");
      __publicField(this, "blockSize", 0);
      __publicField(this, "left", 0);
      __publicField(this, "top", 0);
      __publicField(this, "width", 0);
      __publicField(this, "height", 0);
      __publicField(this, "cellWidth", 0);
      __publicField(this, "cellHeight", 0);
      __publicField(this, "foregroundColorIndex", 0);
      __publicField(this, "backgroundColorIndex", 0);
      __publicField(this, "data", {});
      this.stream = stream;
    }
    parse({ introducer, label }) {
      this.offset = this.stream.getOffset();
      this.blockSize = this.stream.readUint8();
      this.left = this.stream.readUint16();
      this.top = this.stream.readUint16();
      this.width = this.stream.readUint16();
      this.height = this.stream.readUint16();
      this.cellWidth = this.stream.readUint8();
      this.cellHeight = this.stream.readUint8();
      this.foregroundColorIndex = this.stream.readUint8();
      this.backgroundColorIndex = this.stream.readUint8();
      this.data = this.readSubBlocks();
      this.length = this.stream.getOffset() - this.offset;
    }
    export() {
      return {
        type: this.type,
        offset: this.offset,
        length: this.length,
        bytes: getBytes(this.stream, this.offset, this.length),
        data: this.data
      };
    }
  }
  class NetscapeExtension extends BaseExtension {
    constructor(stream) {
      super(stream);
      __publicField(this, "type", "NetscapeExtension");
      __publicField(this, "offset", 0);
      __publicField(this, "length", 0);
      __publicField(this, "bytes", new Uint8Array(0));
      __publicField(this, "blockSize", 0);
      __publicField(this, "subblock", 0);
      __publicField(this, "loop", 0);
      __publicField(this, "terminator", 0);
      __publicField(this, "decode", "");
      this.stream = stream;
    }
    parse({ introducer, label }) {
      this.offset = this.stream.getOffset();
      this.blockSize = this.stream.readUint8();
      this.subblock = this.stream.readUint8();
      this.loop = this.stream.readUint16();
      this.terminator = this.stream.readInt8();
      this.decode = `Loop: ${this.loop > 0 ? this.loop : "forever"}`;
      this.length = this.stream.getOffset() - this.offset;
    }
    export() {
      return {
        type: this.type,
        offset: this.offset,
        length: this.length,
        bytes: getBytes(this.stream, this.offset, this.length),
        data: {
          blockSize: this.blockSize,
          subblock: this.subblock,
          loop: this.loop,
          terminator: this.terminator,
          decode: this.decode
        }
      };
    }
  }
  const _XMP = class {
    constructor(source) {
      __publicField(this, "buffer", null);
      if (source instanceof ArrayBuffer) {
        this.buffer = source;
      } else if (source instanceof Uint8Array) {
        this.buffer = source.buffer;
      } else if (typeof source === "string" || source instanceof String) {
        this.fromDataUri(source);
      }
    }
    static dataUriToBuffer(dataUri) {
      let byteString = atob(dataUri.split(",")[1]), length = byteString.length, out = new ArrayBuffer(length), arr = new Uint8Array(out);
      for (let i = 0; i < length; i++) {
        arr[i] = byteString.charCodeAt(i);
      }
      return out;
    }
    static find(source) {
      let buffer;
      if (source instanceof ArrayBuffer) {
        buffer = source;
      } else if (source instanceof Uint8Array) {
        this.buffer = source.buffer;
      } else if (typeof source === "string" || source instanceof String) {
        buffer = _XMP.dataUriToBuffer(source);
      }
      let view = new DataView(buffer);
      if (!buffer) {
        return;
      }
      const startStr = "<x:xmpmeta", startStrLength = startStr.length, maxStart = buffer.byteLength - startStrLength, endStr = "x:xmpmeta>", endStrLength = endStr.length, maxEnd = buffer.byteLength - endStrLength;
      let start = 2, end = start + startStrLength, found = false;
      while (start < maxStart) {
        if (_XMP.stringFromBuffer(view, start, startStrLength) == startStr) {
          found = true;
          break;
        } else {
          start++;
        }
      }
      if (!found) {
        console.warn("XMP not found");
        return null;
      }
      while (end < maxEnd) {
        if (_XMP.stringFromBuffer(view, end, endStrLength) == endStr) {
          break;
        } else {
          end++;
        }
      }
      end += endStrLength;
      return _XMP.stringFromBuffer(view, start, end - start);
    }
    static stringFromBuffer(buffer, start, length) {
      let out = "";
      for (let i = start; i < start + length; i++) {
        out += String.fromCharCode(buffer.getUint8(i));
      }
      return out;
    }
    static execFind(str, reg) {
      let result = [];
      let isEmpty = true;
      while (isEmpty) {
        let regRxpRxecArray = reg.exec(str);
        if (regRxpRxecArray) {
          result.push(regRxpRxecArray);
        } else {
          isEmpty = false;
        }
      }
      return result;
    }
    static toTree(result) {
      let tree = [];
      if (!result || !result.length) {
        return tree;
      }
      const first = result.shift();
      const firstEndIndex = result.findIndex((item) => {
        return `</${first.tag}>` === item.raw;
      });
      result.splice(firstEndIndex, 1);
      tree.push(__spreadProps(__spreadValues({}, first), {
        children: _XMP.toTree(result)
      }));
      return tree;
    }
    static parse(xmp) {
      if (!xmp) {
        return null;
      }
      const lineReg = new RegExp(`<[^>]+>`, "gi");
      const lineList = _XMP.execFind(xmp, lineReg);
      const out = lineList.map((item, i) => {
        const tagReg = new RegExp("<(.*?)\\s", "g");
        const attrReg = new RegExp('\\s(.*?)="(.*?)"', "g");
        const raw = item[0];
        const tagRes = tagReg.exec(raw);
        const attrExecList = _XMP.execFind(raw, attrReg);
        let attrs = {};
        attrExecList.forEach((element) => {
          const [, key, value] = element;
          attrs[key] = value;
        });
        return {
          raw,
          tag: tagRes ? tagRes[1] : null,
          attrs
        };
      });
      return _XMP.toTree(out);
    }
    fromDataUri(dataUri) {
      this.buffer = _XMP.dataUriToBuffer(dataUri);
    }
    find() {
      return _XMP.find(this.buffer) || "";
    }
    parse() {
      let xmp = this.find();
      return _XMP.parse(xmp);
    }
  };
  let XMP = _XMP;
  __publicField(XMP, "buffer", null);
  class XMPDataExtension extends BaseExtension {
    constructor(stream) {
      super(stream);
      __publicField(this, "type", "XMPDataExtension");
      __publicField(this, "offset", 0);
      __publicField(this, "length", 0);
      __publicField(this, "bytes", new Uint8Array(0));
      __publicField(this, "blockSize", 0);
      __publicField(this, "xmpData", []);
      this.stream = stream;
    }
    parse({ introducer, label }) {
      this.offset = this.stream.getOffset();
      const { blocks } = this.readSubBlocks();
      blocks.forEach((item) => {
        this.blockSize += item.length;
      });
      const arrayBuffer = this.stream.slice(this.offset, this.blockSize);
      let xmp = new XMP(arrayBuffer);
      this.xmpData = xmp.parse();
      this.length = this.stream.getOffset() - this.offset;
    }
    export() {
      return {
        type: this.type,
        offset: this.offset,
        length: this.length,
        bytes: getBytes(this.stream, this.offset, this.length),
        data: {
          blockSize: this.blockSize,
          xmpData: this.xmpData
        }
      };
    }
  }
  class UnknownApplicationExtension extends BaseExtension {
    constructor(stream) {
      super(stream);
      __publicField(this, "type", "UnknownApplicationExtension");
      __publicField(this, "offset", 0);
      __publicField(this, "length", 0);
      __publicField(this, "bytes", new Uint8Array(0));
      __publicField(this, "appData", {});
      this.stream = stream;
    }
    parse({ introducer, label }) {
      this.offset = this.stream.getOffset();
      this.appData = this.readSubBlocks();
      this.length = this.stream.getOffset() - this.offset;
    }
    export() {
      return {
        type: this.type,
        offset: this.offset,
        length: this.length,
        bytes: getBytes(this.stream, this.offset, this.length),
        data: this.appData
      };
    }
  }
  class ApplicationExtension extends BaseExtension {
    constructor(stream) {
      super(stream);
      __publicField(this, "type", "ApplicationExtension");
      __publicField(this, "offset", 0);
      __publicField(this, "length", 0);
      __publicField(this, "bytes", new Uint8Array(0));
      __publicField(this, "introducer", "");
      __publicField(this, "label", "");
      __publicField(this, "blockSize", 0);
      __publicField(this, "identifier", "");
      __publicField(this, "authenticationCode", "");
      __publicField(this, "appData");
      __publicField(this, "netscapeExtension");
      __publicField(this, "xmpdataExtension");
      __publicField(this, "unknownApplicationExtension");
      this.stream = stream;
      this.netscapeExtension = new NetscapeExtension(stream);
      this.xmpdataExtension = new XMPDataExtension(stream);
      this.unknownApplicationExtension = new UnknownApplicationExtension(stream);
      this.appData = {};
    }
    parse({ introducer = 0, label = 0 }) {
      this.introducer = String.fromCharCode(introducer);
      this.label = numberToHex(label);
      this.offset = this.stream.getOffset() - 2;
      this.blockSize = this.stream.readInt8();
      this.identifier = this.stream.readString(8);
      this.authenticationCode = this.stream.readString(3);
      if (this.identifier === "NETSCAPE") {
        this.netscapeExtension.parse({ introducer, label });
        this.appData = this.netscapeExtension.export();
      } else if (this.identifier === "XMP Data") {
        this.xmpdataExtension.parse({ introducer, label });
        this.appData = this.xmpdataExtension.export();
      } else {
        this.unknownApplicationExtension.parse({ introducer, label });
        this.appData = this.unknownApplicationExtension.export();
      }
      this.length = this.stream.getOffset() - this.offset;
    }
    export() {
      return {
        type: this.type,
        offset: this.offset,
        length: this.length,
        bytes: getBytes(this.stream, this.offset, this.length),
        data: {
          introducer: this.introducer,
          label: this.label,
          blockSize: this.blockSize,
          identifier: this.identifier,
          authenticationCode: this.authenticationCode,
          appData: this.appData
        }
      };
    }
  }
  class UnknownExtension extends BaseExtension {
    constructor(stream) {
      super(stream);
      __publicField(this, "type", "UnknownExtension");
      __publicField(this, "data", {});
      this.stream = stream;
    }
    parse({ introducer, label }) {
      this.offset = this.stream.getOffset();
      this.data = this.readSubBlocks();
      this.length = this.stream.getOffset() - this.offset;
    }
    export() {
      return {
        type: this.type,
        offset: this.offset,
        length: this.length,
        bytes: getBytes(this.stream, this.offset, this.length),
        data: this.data
      };
    }
  }
  class Extension {
    constructor(stream) {
      __publicField(this, "extensionBlock", {});
      __publicField(this, "graphicsControlExtension");
      __publicField(this, "plainTextExtension");
      __publicField(this, "commentExtension");
      __publicField(this, "applicationExtension");
      __publicField(this, "unknownExtension");
      this.stream = stream;
      this.stream = stream;
      this.graphicsControlExtension = new GraphicsControlExtension(stream);
      this.commentExtension = new CommentExtension(stream);
      this.plainTextExtension = new PlainTextExtension(stream);
      this.applicationExtension = new ApplicationExtension(stream);
      this.unknownExtension = new UnknownExtension(stream);
    }
    parse({ introducer }) {
      const label = this.stream.readUint8();
      switch (label) {
        case 249:
          {
            this.graphicsControlExtension.parse({ introducer, label });
            this.extensionBlock = this.graphicsControlExtension.export();
          }
          break;
        case 254:
          this.commentExtension.parse({ introducer, label });
          this.extensionBlock = this.commentExtension.export();
          break;
        case 1:
          this.plainTextExtension.parse({ introducer, label });
          this.extensionBlock = this.plainTextExtension.export();
          break;
        case 255:
          {
            this.applicationExtension.parse({ introducer, label });
            this.extensionBlock = this.applicationExtension.export();
          }
          break;
        default:
          this.unknownExtension.parse({ introducer, label });
          this.extensionBlock = this.unknownExtension.export();
          break;
      }
    }
    export() {
      return this.extensionBlock;
    }
  }
  class LocalColorTable extends ColorTable {
    constructor(stream) {
      super(stream, "LocalColorTable");
    }
  }
  class ImageDescriptor {
    constructor(stream) {
      __publicField(this, "type", "ImageDescriptor");
      __publicField(this, "offset", 0);
      __publicField(this, "length", 0);
      __publicField(this, "bytes", new Uint8Array(0));
      __publicField(this, "introducer", "");
      __publicField(this, "left", 0);
      __publicField(this, "top", 0);
      __publicField(this, "width", 0);
      __publicField(this, "height", 0);
      __publicField(this, "packedFields", []);
      __publicField(this, "localColorTableFlag", 0);
      __publicField(this, "interlaceFlag", 0);
      __publicField(this, "sortFlag", 0);
      __publicField(this, "reserved", []);
      __publicField(this, "localColorTableSize", 0);
      this.stream = stream;
      this.stream = stream;
    }
    hasLocalColorTable() {
      return !!this.localColorTableFlag;
    }
    getLocalColorTableSize() {
      return this.localColorTableSize;
    }
    parse({ introducer = 0 }) {
      this.introducer = String.fromCharCode(introducer);
      this.offset = this.stream.getOffset() - 1;
      this.left = this.stream.readUint16();
      this.top = this.stream.readUint16();
      this.width = this.stream.readUint16();
      this.height = this.stream.readUint16();
      this.packedFields = byteToBits(this.stream.readUint8());
      const bits = [...this.packedFields];
      this.localColorTableFlag = bits.shift() || 0;
      this.interlaceFlag = bits.shift() || 0;
      this.sortFlag = bits.shift() || 0;
      this.reserved = bits.splice(0, 2);
      this.localColorTableSize = bitsToNumber(bits.splice(0, 3));
      this.length = this.stream.getOffset() - this.offset;
    }
    export() {
      return {
        type: this.type,
        offset: this.offset,
        length: this.length,
        bytes: getBytes(this.stream, this.offset, this.length),
        data: {
          introducer: this.introducer,
          left: this.left,
          top: this.top,
          width: this.width,
          height: this.height,
          packedFields: this.packedFields,
          localColorTableFlag: this.localColorTableFlag,
          interlaceFlag: this.interlaceFlag,
          sortFlag: this.sortFlag,
          reserved: this.reserved,
          localColorTableSize: this.localColorTableSize
        }
      };
    }
  }
  class BitReader {
    constructor(bytes) {
      __publicField(this, "bytes");
      __publicField(this, "byteOffset");
      __publicField(this, "bitOffset");
      __publicField(this, "totalByteOffset");
      this.bytes = bytes || new Int8Array();
      this.byteOffset = 0;
      this.bitOffset = 0;
      this.totalByteOffset = 0;
    }
    readBits(len) {
      let result = 0;
      let rbits = 0;
      while (rbits < len) {
        if (this.byteOffset >= this.bytes.length) {
          throw new Error(`Not enough bytes to read ${len} bits (read ${rbits} bits)`);
        }
        const bbits = Math.min(8 - this.bitOffset, len - rbits);
        const mask = 255 >> 8 - bbits << this.bitOffset;
        result += (this.bytes[this.byteOffset] & mask) >> this.bitOffset << rbits;
        rbits += bbits;
        this.bitOffset += bbits;
        if (this.bitOffset === 8) {
          this.byteOffset += 1;
          this.totalByteOffset += 1;
          this.bitOffset = 0;
        }
      }
      return result;
    }
    hasBits(len = 1) {
      if (len > 12) {
        throw new Error(`Exceeds max bit size: ${len} (max: 12)`);
      }
      if (this.byteOffset >= this.bytes.length)
        return false;
      const bitsRemain = 8 - this.bitOffset;
      if (len <= bitsRemain)
        return true;
      const bytesRemain = this.bytes.length - this.byteOffset - 1;
      if (bytesRemain < 1)
        return false;
      if (len > bitsRemain + 8 * bytesRemain)
        return false;
      return true;
    }
    setBytes(bytes, byteOffset = 0, bitOffset = 0) {
      this.bytes = bytes;
      this.byteOffset = byteOffset;
      this.bitOffset = bitOffset;
    }
    pushBytes(bytes) {
      if (this.hasBits()) {
        const remainBytes = this.bytes.length - this.byteOffset;
        const extended = new Uint8Array(remainBytes + bytes.length);
        extended.set(this.bytes.slice(this.byteOffset));
        extended.set(bytes, remainBytes);
        this.bytes = extended;
        this.byteOffset = 0;
      } else {
        this.bytes = bytes;
        this.byteOffset = 0;
        this.bitOffset = 0;
      }
    }
    getState() {
      return {
        bitOffset: this.bitOffset,
        byteOffset: this.totalByteOffset
      };
    }
  }
  class ImageContent {
    constructor(stream) {
      __publicField(this, "type", "ImageContent");
      __publicField(this, "offset", 0);
      __publicField(this, "length", 0);
      __publicField(this, "bytes", new Uint8Array(0));
      __publicField(this, "children", {});
      this.stream = stream;
      this.stream = stream;
    }
    parse() {
      this.offset = this.stream.getOffset();
      const lzwData = this.readImageContent();
      this.children = this.lzwDecode(lzwData);
      this.length = this.stream.getOffset() - this.offset;
    }
    lzwDecode(data) {
      const { blocks, lzwMinimumCodeSize } = data;
      const indexStream = [];
      const codeUnits = [];
      let codeStream = [];
      let codeTable = [];
      const br = new BitReader();
      const clearCode = 2 << lzwMinimumCodeSize - 1;
      const eoiCode = clearCode + 1;
      let lastCode = eoiCode;
      let size = lzwMinimumCodeSize + 1;
      let growCode = (2 << size - 1) - 1;
      let isInitialized = false;
      for (const { offset, length } of blocks) {
        br.pushBytes(this.stream.slice(offset, length));
        while (br.hasBits(size)) {
          const codeStart = br.getState();
          const code = br.readBits(size);
          if (code === eoiCode) {
            codeStream.push(code);
            break;
          } else if (code === clearCode) {
            codeUnits.push({ stream: [], table: [], start: codeStart });
            codeStream = codeUnits[codeUnits.length - 1].stream;
            codeTable = codeUnits[codeUnits.length - 1].table;
            for (let i = 0; i <= eoiCode; i++) {
              codeTable[i] = i < clearCode ? [i] : [];
            }
            lastCode = eoiCode;
            size = lzwMinimumCodeSize + 1;
            growCode = (2 << size - 1) - 1;
            isInitialized = false;
          } else if (!isInitialized) {
            indexStream.push(...codeTable[code]);
            isInitialized = true;
          } else {
            let k = 0;
            const prevCode = codeStream[codeStream.length - 1];
            if (code <= lastCode) {
              indexStream.push(...codeTable[code]);
              k = codeTable[code][0];
            } else {
              k = codeTable[prevCode][0];
              indexStream.push(...codeTable[prevCode], k);
            }
            if (lastCode < 4095) {
              lastCode += 1;
              codeTable[lastCode] = [...codeTable[prevCode], k];
              if (lastCode === growCode && lastCode < 4095) {
                size += 1;
                growCode = (2 << size - 1) - 1;
              }
            }
          }
          codeStream.push(code);
        }
      }
      return Object.assign({}, data, {
        indexStream,
        clearCode,
        eoiCode,
        codeUnits,
        blockCount: blocks.length,
        codeUnitCount: codeUnits.length
      });
    }
    export() {
      return {
        type: this.type,
        offset: this.offset,
        length: this.length,
        bytes: getBytes(this.stream, this.offset, this.length),
        data: this.children
      };
    }
    readImageContent() {
      const offset = this.stream.getOffset();
      const lzwMinimumCodeSize = this.stream.readUint8();
      let { blocks } = this.readSubBlocks();
      const length = this.stream.getOffset() - offset;
      return {
        offset,
        length,
        type: "image",
        lzwMinimumCodeSize,
        blocks
      };
    }
    readSubBlocks() {
      const offset = this.stream.getOffset();
      let blockSize = this.stream.readUint8();
      let length = 1;
      const blocks = [];
      while (blockSize > 0) {
        blocks.push({ offset: offset + length, length: blockSize });
        length += blockSize + 1;
        this.stream.seek(blockSize);
        blockSize = this.stream.readUint8();
      }
      length += 1;
      return {
        blocks,
        blocksLength: length
      };
    }
  }
  class ImageData {
    constructor(stream) {
      __publicField(this, "localColorTable");
      __publicField(this, "imageDescriptor");
      __publicField(this, "imageContent");
      this.stream = stream;
      this.stream = stream;
      this.localColorTable = new LocalColorTable(stream);
      this.imageDescriptor = new ImageDescriptor(stream);
      this.imageContent = new ImageContent(stream);
    }
    parse({ introducer }) {
      this.imageDescriptor.parse({ introducer });
      if (this.imageDescriptor.hasLocalColorTable()) {
        const localColorTableSize = this.imageDescriptor.getLocalColorTableSize();
        this.localColorTable.parse({ size: localColorTableSize });
      }
      this.imageContent.parse();
    }
    createFrame(colorTable, cidx, width, height) {
      const canvas = document.createElement("canvas");
      canvas.width = width;
      canvas.height = height;
      const context = canvas.getContext("2d");
      const imgData = context.createImageData(width, height);
      const pixels = imgData.data;
      for (let i = 0, poff = 0; i < cidx.length; i += 1, poff += 4) {
        pixels[poff + 0] = colorTable[cidx[i]][0];
        pixels[poff + 1] = colorTable[cidx[i]][1];
        pixels[poff + 2] = colorTable[cidx[i]][2];
        pixels[poff + 3] = cidx[i] !== 88 ? 255 : 0;
      }
      return imgData;
    }
    exportFrameData(globalColorTableExportData) {
      const { width, height, left, top, localColorTableFlag } = this.imageDescriptor.export().data;
      const { indexStream } = this.imageContent.export().data;
      let colorTableData = this.imageDescriptor.hasLocalColorTable() ? this.localColorTable.export() : globalColorTableExportData || {};
      return {
        type: "Frame",
        data: {
          width,
          height,
          left,
          top,
          localColorTableFlag,
          imageData: this.createFrame(colorTableData.data, indexStream, width, height)
        }
      };
    }
    export() {
      const id = this.imageDescriptor.export();
      const ic = this.imageContent.export();
      if (!this.imageDescriptor.hasLocalColorTable()) {
        return [id, ic];
      }
      const lct = this.localColorTable.export();
      return [id, lct, ic];
    }
  }
  class Block {
    constructor(stream) {
      __publicField(this, "type", "Block");
      __publicField(this, "offset", 0);
      __publicField(this, "length", 0);
      __publicField(this, "blocks", []);
      __publicField(this, "extension");
      __publicField(this, "imageData");
      this.stream = stream;
      this.stream = stream;
      this.extension = new Extension(this.stream);
      this.imageData = new ImageData(this.stream);
    }
    parse(globalColorTableExportData) {
      this.offset = this.stream.getOffset();
      while (this.stream.hasMore()) {
        const introducer = this.stream.readUint8();
        switch (String.fromCharCode(introducer)) {
          case "!":
            this.extension.parse({ introducer });
            this.blocks.push(this.extension.export());
            break;
          case ",":
            this.imageData.parse({ introducer });
            this.blocks.push(...this.imageData.export());
            break;
          case ";":
            break;
          default:
            throw new Error("Unknown block: 0x" + introducer.toString(16));
        }
      }
      this.length = this.stream.getOffset() - this.offset;
    }
    export() {
      return this.blocks;
    }
  }
  class Parser {
    constructor(arrayBuffer) {
      __publicField(this, "parsed");
      __publicField(this, "blockList");
      __publicField(this, "dataList");
      __publicField(this, "stream");
      __publicField(this, "header");
      __publicField(this, "logicalScreenDescriptor");
      __publicField(this, "globalColorTable");
      __publicField(this, "block");
      __publicField(this, "isExport", false);
      this.parsed = false;
      this.blockList = [];
      this.dataList = [];
      this.stream = new Stream(arrayBuffer, true);
      this.header = new Header(this.stream);
      this.logicalScreenDescriptor = new LogicalScreenDescriptor(this.stream);
      this.globalColorTable = new GlobalColorTable(this.stream);
      this.block = new Block(this.stream);
      this.parse();
    }
    parse() {
      if (this.parsed) {
        return;
      }
      this.header.parse();
      this.logicalScreenDescriptor.parse();
      if (this.logicalScreenDescriptor.hasGlobalColorTable()) {
        const size = this.logicalScreenDescriptor.getGlobalColorTableSize();
        this.globalColorTable.parse({ size });
      }
      this.block.parse(this.logicalScreenDescriptor.hasGlobalColorTable() ? this.globalColorTable.export() : void 0);
      this.parsed = true;
    }
    getSize() {
      const { data } = this.logicalScreenDescriptor.export();
      return [data.width, data.height];
    }
    export() {
      if (this.isExport) {
        return;
      }
      this.blockList.push(this.header.export());
      this.blockList.push(this.logicalScreenDescriptor.export());
      if (this.logicalScreenDescriptor.hasGlobalColorTable()) {
        this.blockList.push(this.globalColorTable.export());
      }
      this.blockList.push(...this.block.export());
      this.dataHandler();
      this.isExport = true;
    }
    exportWarn() {
      console.warn.call(this, `parser.export()`);
      return [];
    }
    getBlockList() {
      return this.blockList;
    }
    getDataList() {
      if (!this.isExport) {
        return this.exportWarn();
      }
      return this.dataList;
    }
    createImageData({ colorTable, indexStream, width, height, transparency }) {
      const canvas = document.createElement("canvas");
      canvas.width = width;
      canvas.height = height;
      const context = canvas.getContext("2d");
      const imageData = context.createImageData(width, height);
      const pixels = imageData.data;
      for (let i = 0, poff = 0; i < indexStream.length; i += 1, poff += 4) {
        if (indexStream[i] !== transparency) {
          pixels[poff + 0] = colorTable[indexStream[i]][0];
          pixels[poff + 1] = colorTable[indexStream[i]][1];
          pixels[poff + 2] = colorTable[indexStream[i]][2];
          pixels[poff + 3] = 255;
        } else {
          pixels[poff + 0] = 0;
          pixels[poff + 1] = 0;
          pixels[poff + 2] = 0;
          pixels[poff + 3] = 0;
        }
      }
      return imageData;
    }
    dataHandler() {
      let frameData = {};
      const globalColorTable = this.blockList.find((item) => item.type === "GlobalColorTable");
      this.blockList.forEach((item) => {
        if (item.type === "GraphicsControlExtension") {
          frameData[item.type] = item.data;
          this.dataList.push(item);
        } else if (item.type === "ImageDescriptor") {
          frameData[item.type] = item.data;
          this.dataList.push(item);
        } else if (item.type === "ImageContent") {
          frameData[item.type] = item.data;
          this.dataList.push(item);
          const { delayTime, disposalMethod, transparentColorFlag, transparentColorIndex } = frameData.GraphicsControlExtension;
          const { localColorTableFlag, left, top, width, height } = frameData.ImageDescriptor;
          const { indexStream } = frameData.ImageContent;
          const transparency = transparentColorFlag ? transparentColorIndex : null;
          const colorTable = localColorTableFlag ? null : globalColorTable == null ? void 0 : globalColorTable.data;
          frameData["Frame"] = {
            delayTime,
            disposalMethod,
            transparency,
            colorTable,
            indexStream,
            left,
            top,
            width,
            height
          };
          Object.assign(frameData["Frame"], {
            imageData: this.createImageData(frameData["Frame"])
          });
          this.dataList.push({
            type: "Frame",
            data: frameData["Frame"]
          });
        } else {
          this.dataList.push(item);
        }
      });
    }
    getFrames() {
      if (!this.isExport) {
        return this.exportWarn();
      }
      return this.dataList.filter((item) => item.type === "Frame").map((item) => {
        const { delayTime, transparency, left, top, width, height, imageData } = item.data;
        return {
          delayTime,
          transparency,
          left,
          top,
          width,
          height,
          imageData
        };
      });
    }
    getBlockItem(type) {
      const currentItem = this.blockList.find((item) => item.type === type);
      return currentItem && currentItem.data;
    }
  }

  let canvas, offscreenCanvas;
  function getCanvas() {
      if (canvas) {
          return canvas;
      }
      canvas = document.createElement('canvas');
      return canvas;
  }

  function getOffscreenCanvas() {
      if (offscreenCanvas) {
          return offscreenCanvas;
      }
      if (OffscreenCanvas) {
          offscreenCanvas = new OffscreenCanvas(10, 10);
      }
      return offscreenCanvas;
  }

  function putImageData(canvas, frame, gif) {
      const { maxWidth, maxHeight } = gif;
      canvas.width = maxWidth;
      canvas.height = maxHeight;
      const ctx = canvas.getContext('2d');
      ctx.clearRect(0, 0, canvas.width, canvas.height);
      const { imageData, left, top } = frame;
      ctx.putImageData(imageData, left, top);
      frame.maxWidth = maxWidth;
      frame.maxHeight = maxHeight;
  }

  const gifCollection = [];

  function frameLoop(timeStamp) {
      gifCollection.forEach(gif => {
          const frames = gif.frames;
          if (!gif.loaded || !gif._playing || !frames || frames.length === 0) {
              return;
          }
          if (gif._idx === frames.length) {
              gif._idx = 0;
          }
          const frame = frames[gif._idx];
          frame.index = gif._idx;
          if (!frame.dataURL) {
              const canvas = getCanvas();
              putImageData(canvas, frame, gif);
              frame.dataURL = canvas.toDataURL();

              const offCanvas = getOffscreenCanvas();
              if (offCanvas) {
                  putImageData(offCanvas, frame, gif);
                  frame.imageBitMap = offCanvas.transferToImageBitmap();
              }
          }
          if (!gif._time) {
              gif._time = timeStamp;
              gif._idx++;
              gif._mitt.emit('frame', frame);
          } else {
              if (timeStamp - gif._time >= gif.options.loopTime) {
                  gif._time = timeStamp;
                  gif._idx++;
                  gif._mitt.emit('frame', frame);
              }
          }
      });
      requestAnimationFrame(frameLoop);
  }
  setTimeout(() => {
      requestAnimationFrame(frameLoop);
  }, 16);

  class GIF {
      constructor(options) {
          this.options = Object.assign({}, { loopTime: 70 }, options);
          this.options.loopTime = Math.max(17, this.options.loopTime);
          this._mitt = mitt();
          this._idx = 0;
          this._playing = true;
          this.loaded = false;
          this.frames = [];
          if (!options.url) {
              console.error('not find gif url');
          } else {
              fetch(options.url).then(res => res.arrayBuffer()).then(arrayBuffer => {
                  const parser = new Parser(arrayBuffer);
                  parser.export();
                  this.frames = parser.getFrames();
                  let maxWidth = 0, maxHeight = 0;
                  this.frames.forEach(frame => {
                      const { width, height, left, top } = frame;
                      maxWidth = Math.max(maxWidth, width + left);
                      maxHeight = Math.max(maxHeight, height + top);
                  });
                  this.maxWidth = maxWidth;
                  this.maxHeight = maxHeight;
                  gifCollection.push(this);
                  this.loaded = true;
              }).catch(error => {
                  console.error(error);
              });
          }
      }

      config(options = {}) {
          this.options = Object.assign(this.options, options);
          return this;
      }

      on(eventType, handler) {
          this._mitt.on(eventType, handler);
          return this;
      }

      off(eventType, handler) {
          this._mitt.off(eventType, handler);
          return this;
      }

      play() {
          this._playing = true;
          return this;
      }

      pause() {
          this._playing = false;
          return this;
      }

      isPlay() {
          return this._playing;
      }

      dispose() {
          const index = gifCollection.indexOf(this);
          if (index > -1) {
              gifCollection.splice(index, 1);
          }
          this._mitt.all.clear();
          this.frames = null;
          return this;
      }
  }

  exports.GIF = GIF;

  Object.defineProperty(exports, '__esModule', { value: true });

}));
//# sourceMappingURL=gifloop.js.map
